#include "pch.h"
#include "UnderlineFormatter.h"

using namespace D2D1;
using namespace Microsoft::WRL;

// Constructor
UnderlineFormatter::UnderlineFormatter() :
    m_refCount(0),
    m_dpiTransform(Matrix3x2F::Identity()),
    m_renderTransform(Matrix3x2F::Identity()),
    m_worldToPixel(Matrix3x2F::Identity()),
    m_pixelToWorld(Matrix3x2F::Identity())
{
}

// IUnknown methods
ULONG STDMETHODCALLTYPE UnderlineFormatter::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE UnderlineFormatter::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE UnderlineFormatter::QueryInterface(_In_ REFIID riid,
                                                             _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(IDWriteTextRenderer))
    {
        *ppOutput = static_cast<IDWriteTextRenderer*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IDWritePixelSnapping))
    {
        *ppOutput = static_cast<IDWritePixelSnapping*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

// IDWritePixelSnapping methods
HRESULT UnderlineFormatter::IsPixelSnappingDisabled(void * clientDrawingContext,
    _Out_ BOOL * isDisabled)
{
    *isDisabled = false;
    return S_OK;
}

HRESULT UnderlineFormatter::GetPixelsPerDip(void * clientDrawingContext,
                                            _Out_ FLOAT * pixelsPerDip)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    float dpiX, dpiY;
    drawingContext->renderTarget->GetDpi(&dpiX, &dpiY);
    *pixelsPerDip = dpiX / 96;

    // Save DPI as transform for pixel snapping
    m_dpiTransform = Matrix3x2F::Scale(dpiX / 96.0f, dpiY / 96.0f);
    m_worldToPixel = m_renderTransform * m_dpiTransform;
    m_pixelToWorld = m_worldToPixel;
    m_pixelToWorld.Invert();

    return S_OK;
}

HRESULT UnderlineFormatter::GetCurrentTransform(void * clientDrawingContext,
                                                DWRITE_MATRIX * transform)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    // Matrix structures are defined identically
    drawingContext->renderTarget->GetTransform((D2D1_MATRIX_3X2_F *) transform);

    // Save transform for pixel snapping
    m_renderTransform = *(Matrix3x2F *) transform;
    m_worldToPixel = m_renderTransform * m_dpiTransform;
    m_pixelToWorld = m_worldToPixel;
    m_pixelToWorld.Invert();

    return S_OK;
}

// IDWriteTextRenderer methods
HRESULT UnderlineFormatter::DrawGlyphRun(void * clientDrawingContext,
                                         FLOAT baselineOriginX,
                                         FLOAT baselineOriginY,
                                         DWRITE_MEASURING_MODE measuringMode,
                                         _In_ const DWRITE_GLYPH_RUN * glyphRun,
                                         _In_ const DWRITE_GLYPH_RUN_DESCRIPTION *
                                             glyphRunDescription,
                                         IUnknown * clientDrawingEffect)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    ID2D1RenderTarget * renderTarget = drawingContext->renderTarget;

    // Get brush and underline specifier
    ID2D1Brush * brush = drawingContext->defaultBrush;
    UnderlineSpecifier * underlineSpecifier = nullptr;
    GetTextDrawingEffects(clientDrawingEffect, &brush, &underlineSpecifier);

    renderTarget->DrawGlyphRun(Point2F(baselineOriginX, baselineOriginY),
                               glyphRun, brush, measuringMode);

    return S_OK;
}

HRESULT UnderlineFormatter::DrawUnderline(void * clientDrawingContext,
                                          FLOAT baselineOriginX,
                                          FLOAT baselineOriginY,
                                          _In_ const DWRITE_UNDERLINE *
                                              underline,
                                          IUnknown * clientDrawingEffect)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    // Get brush and underline specifier
    ID2D1Brush * brush = drawingContext->defaultBrush;
    UnderlineSpecifier * underlineSpecifier = nullptr;
    GetTextDrawingEffects(clientDrawingEffect, &brush, &underlineSpecifier);

    // Get underline count
    int underlineCount = 1;

    if (underlineSpecifier != nullptr)
        underlineCount = underlineSpecifier->GetUnderlineCount();

    if (underlineCount < 1 || underlineCount > 3)
        return E_INVALIDARG;

    if (underlineCount == 1 || underlineCount == 3)
    {
        FillRectangle(drawingContext->renderTarget,
                      brush,
                      baselineOriginX,
                      baselineOriginY + underline->offset,
                      underline->width,
                      underline->thickness);
    }

    if (underlineCount == 2 || underlineCount == 3)
    {
        FillRectangle(drawingContext->renderTarget,
                      brush,
                      baselineOriginX,
                      baselineOriginY + underline->offset + 
                          (underlineCount - 1) * underline->thickness,
                      underline->width,
                      underline->thickness);

        FillRectangle(drawingContext->renderTarget,
                      brush,
                      baselineOriginX,
                      baselineOriginY + underline->offset - 
                          (underlineCount - 1) * underline->thickness,
                      underline->width,
                      underline->thickness);
    }

    return S_OK;
}

HRESULT UnderlineFormatter::DrawStrikethrough(void * clientDrawingContext,
                                              FLOAT baselineOriginX,
                                              FLOAT baselineOriginY,
                                              _In_ const DWRITE_STRIKETHROUGH *
                                                  strikethrough,
                                              IUnknown * clientDrawingEffect)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    // Get brush and underline specifier
    ID2D1Brush * brush = drawingContext->defaultBrush;
    UnderlineSpecifier * underlineSpecifier = nullptr;
    GetTextDrawingEffects(clientDrawingEffect, &brush, &underlineSpecifier);

    // Get strikeout count
    int strikethroughCount = 1;

    if (underlineSpecifier != nullptr)
        strikethroughCount = underlineSpecifier->GetStrikethroughCount();

    if (strikethroughCount < 1 || strikethroughCount > 3)
        return E_INVALIDARG;

    if (strikethroughCount == 1 || strikethroughCount == 3)
    {
        FillRectangle(drawingContext->renderTarget,
                      brush,
                      baselineOriginX,
                      baselineOriginY + strikethrough->offset,
                      strikethrough->width,
                      strikethrough->thickness);
    }
    if (strikethroughCount == 2 || strikethroughCount == 3)
    {
        FillRectangle(drawingContext->renderTarget,
                      brush,
                      baselineOriginX,
                      baselineOriginY + strikethrough->offset + 
                          (strikethroughCount - 1) * strikethrough->thickness,
                      strikethrough->width,
                      strikethrough->thickness);

        FillRectangle(drawingContext->renderTarget,
                      brush,
                      baselineOriginX,
                      baselineOriginY + strikethrough->offset - 
                          (strikethroughCount - 1) * strikethrough->thickness,
                      strikethrough->width,
                      strikethrough->thickness);
    }
    return S_OK;
}

HRESULT UnderlineFormatter::DrawInlineObject(void * clientDrawingContext,
                                             FLOAT originX,
                                             FLOAT originY,
                                             IDWriteInlineObject * inlineObject,
                                             BOOL isSideways,
                                             BOOL isRightToLeft,
                                             IUnknown * clientDrawingEffect)
{
    DrawingContext * drawingContext =
        static_cast<DrawingContext *>(clientDrawingContext);

    return inlineObject->Draw(clientDrawingContext,
                              this,
                              originX,
                              originY,
                              isSideways,
                              isRightToLeft,
                              clientDrawingEffect);
}

void UnderlineFormatter::GetTextDrawingEffects(IUnknown * clientDrawingEffect,
                                               ID2D1Brush ** pBrush,
                                               UnderlineSpecifier ** pUnderlineSpecifier)
{
    if (clientDrawingEffect != nullptr)
    {
        void * pInterface;

        if (S_OK == clientDrawingEffect->QueryInterface(__uuidof(ID2D1Brush), 
                                                        &pInterface))
            *pBrush = (ID2D1Brush *) pInterface;

        else
            *pUnderlineSpecifier = (UnderlineSpecifier *) clientDrawingEffect;
    }
}

void UnderlineFormatter::FillRectangle(ID2D1RenderTarget * renderTarget,
                                       ID2D1Brush * brush,
                                       float x, float y,
                                       float width, float thickness)
{
    // Snap the y coordinate to the nearest pixel
    D2D1_POINT_2F pt = Point2F(0, y);
    pt = m_worldToPixel.TransformPoint(pt);
    pt.y = (float) (int) (pt.y + 0.5f);
    pt = m_pixelToWorld.TransformPoint(pt);
    y = pt.y;

    // Fill the rectangle
    D2D1_RECT_F rect = RectF(x, y, x + width, y + thickness);
    renderTarget->FillRectangle(&rect, brush);
}
